/*
 *  Copyright (C) 2015  Malte Humann
 *
 *  This file is part of the Time Series Anomaly Detection (TSAD).
 *
 *  TSAD is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  TSAD is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with TSAD.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.tzi.imonitor;

import java.io.IOException;
import java.lang.Thread.UncaughtExceptionHandler;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.sql.Statement;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * The main class of the time series anomaly detection (TSAD).
 * <p>
 * This class manages the other components of the TSAD and is therefore
 * responsible for providing a database connection for and "communication"
 * between those components as well as for starting and stopping them.
 */
public class AnomalyDetection {
	/** The check time series component. */
	private static CheckTimeSeries check;
	/** The spool files reader component. */
	private static SpoolFilesReader reader;
	/** The spool files watch dog component. */
	private static SpoolFilesWatchDog watchdog;
	/** The SQL database connection used by all components. */
	private static Connection conn = null;
	private final static Logger LOGGER = LoggerFactory
			.getLogger(AnomalyDetection.class);

	/**
	 * Sets up the database connection and initialize the database if necessary.
	 * Afterwards creates the individual components and starts them.
	 * <p>
	 * Also provides an uncaught exception handler that forces all components to
	 * {@link #stop(Throwable) stop}.
	 * 
	 * @param args
	 *            unused
	 */
	public static void main(String[] args) {
		Path spooldir = Paths.get(Configuration.getProperty("spool_dir"));
		// TODO check if relevant resources are available, e.g. R, send_nsca
		// (including configuration)

		try {
			Class.forName("org.sqlite.JDBC");
		} catch (ClassNotFoundException e) {
			LOGGER.error("Unable to load SQLite JDBC driver.", e);
			return;
		}

		try {
			conn = DriverManager.getConnection(Configuration
					.getProperty("jdbc_url"));
		} catch (SQLException e1) {
			LOGGER.error("Unable to connect to database.", e1);
			if (conn != null) {
				try {
					conn.close();
				} catch (SQLException e2) {
					LOGGER.warn("Unable to close database connection.", e2);
				}
			}
			return;
		}

		try (Statement stmt = conn.createStatement()) {
			stmt.executeUpdate("CREATE TABLE IF NOT EXISTS hostnames("
					+ "id INTEGER PRIMARY KEY ASC AUTOINCREMENT, "
					+ "hostname TEXT UNIQUE NOT NULL)");
			stmt.executeUpdate("CREATE TABLE IF NOT EXISTS servicedescriptions("
					+ "id INTEGER PRIMARY KEY ASC AUTOINCREMENT, "
					+ "servicedescription TEXT UNIQUE NOT NULL)");
			stmt.executeUpdate("CREATE TABLE IF NOT EXISTS perfdatalabels("
					+ "id INTEGER PRIMARY KEY ASC AUTOINCREMENT, "
					+ "perfdatalabel TEXT UNIQUE NOT NULL)");
			stmt.executeUpdate("CREATE TABLE IF NOT EXISTS timeseries("
					+ "id INTEGER PRIMARY KEY ASC AUTOINCREMENT, "
					+ "hostname_id INTEGER, "
					+ "servicedescription_id INTEGER, "
					+ "perfdatalabel_id INTEGER, "
					+ "lastprocessed INTEGER, "
					+ "estimated_period INTEGER, "
					+ "FOREIGN KEY (hostname_id) REFERENCES hostnames(id), "
					+ "FOREIGN KEY (servicedescription_id) REFERENCES servicedescriptions(id), "
					+ "FOREIGN KEY (perfdatalabel_id) REFERENCES perfdatalabels(id), "
					+ "UNIQUE (hostname_id, servicedescription_id, perfdatalabel_id))");
			// TODO adjusted time and value for constant interval lengths?
			// TODO dynamic programming related data storage?
			stmt.executeUpdate("CREATE TABLE IF NOT EXISTS perfdata("
					+ "timestamp INTEGER, " + "timeseries_id INTEGER, "
					+ "value REAL, predicted REAL, score BOOLEAN, state INT, "
					+ "FOREIGN KEY (timeseries_id) REFERENCES timeseries(id), "
					+ "PRIMARY KEY (timestamp, timeseries_id))");
		} catch (SQLException e) {
			LOGGER.error("Unable to set up database tables.", e);
			return;
		}

		check = new CheckTimeSeries(conn);
		reader = new SpoolFilesReader(check, conn);
		try {
			watchdog = new SpoolFilesWatchDog(reader, spooldir);
		} catch (IOException e) {
			LOGGER.error("Unable to create spool directory watcher.", e);
			return;
		}

		Thread.setDefaultUncaughtExceptionHandler(new UncaughtExceptionHandler() {
			@Override
			public void uncaughtException(Thread t, Throwable e) {
				stop(e);
			}
		});

		// start all of them
		check.start();
		reader.start();
		watchdog.start();
	}

	/**
	 * Issues a stop command for all components with no further explanation.
	 * <p>
	 * Use in case of a safe shutdown.
	 * 
	 * @see #stop(Throwable)
	 */
	public static void stop() {
		stop(null);
	}

	/**
	 * Issues a stop command for all components including a reason.
	 * <p>
	 * Use in case of an unexpected shutdown.
	 * 
	 * @param reason
	 *            the cause of the shutdown
	 * @see #stop()
	 */
	public static void stop(Throwable reason) {
		if (reason == null) {
			LOGGER.debug("Stopping.");
		} else {
			LOGGER.warn("Unexpected shut down.", reason);
		}
		check.stop();
		reader.stop();
		watchdog.stop();
		if (conn != null) {
			try {
				conn.close();
			} catch (SQLException e) {
				LOGGER.warn("Unable to close database connection.", e);
			}
		}
	}
}

